"""Interfaces and classes for Moving Targets Masters."""

from moving_targets.masters import losses, optimizers, backends
from moving_targets.masters.masters import Master, RegressionMaster, ClassificationMaster
